"""
A package for analyzing student code.

TODO: Statistics
TODO: Output checking for folder of submissions
TODO: Handle tabular output for unit tests
TODO: Big annotated example
TODO: Fix sections example
TODO: Set up "Unit Test" environment
TODO: Add documentation for each one, particularly toolkit
TODO: Set up command line and other environments
TODO: Improve feedback tracking for all the tools
TODO: Fix up imports everywhere to be consistent, particularly for the modules themselves
TODO: A/B Tool - override existing feedback, and tools for randomly or deterministically choosing students

DONE: Rework Source module
DONE: Rework MAIN_REPORT keyword arg for all tools
"""

# Core Commands
from pedal.core.commands import *
from pedal.source import *
