class Tag:
    """

    """
    def __init__(self, kind, name, explanation, weight):
        self.kind = kind
        self.name = name
        self.explanation = explanation
        self.weight = weight
