__all__ = ['Resolver']


class Resolver:
    def resolve(self):
        """

        """
        pass

    def _on_pre_resolve(self):
        """
        Hook event.
        Returns:

        """

    def _on_post_resolve(self):
        """
        Hook event.
        Returns:

        """
